;
; Slip: Lisp in 100 lines - Theo D'Hondt: SOFT: VUB - 2010
;
;       CPS* version
;
;       Version 2 - a native call-cc
;                   
; <expression>  ::= <computation>|<lambda>|<quote>|<variable>|
;                   <literal>|<null> 
; <computation> ::= <definition>|<assignment>|<sequence>|
;                   <conditional>|<iteration>|<application>
; <definition>  ::= (define <variable> <expression>)
; <definition>  ::= (define <pattern> <expression>+)
; <assignment>  ::= (set! <variable> <expression>)
; <sequence>    ::= (begin <expression>+)
; <conditional> ::= (if <expression> <expression> <expression>)
; <conditional> ::= (if <expression> <expression>)
; <iteration>   ::= (while <expression> <expression>+)
; <application> ::= (<expression>+)
; <lambda>      ::= (lambda () <expression>+)
; <lambda>      ::= (lambda <variable> <expression>+)
; <lambda>      ::= (lambda (<pattern>) <expression>+)
; <quote>       ::= '[s-expression]
; <variable>    ::= [symbol]
; <pattern>     ::= (<variable>+)
; <pattern>     ::= (<variable>+ . <variable>)
; <literal>     ::= [number]|[character]|[string]|#t|#f
; <null>        ::= ()

; don't use syntactic constructs (e.g. cond, do, ...) 

(begin
  (define meta-level-eval eval)

;
; natives
;

  (define (wrap-native-procedure native-procedure)
    (lambda (arguments continue environment)
      (define native-value (apply native-procedure arguments))
      (continue native-value environment)))

  (define (cps-apply expression continue environment)
    (define procedure (car expression))
    (define arguments (cadr expression))
    (procedure arguments continue environment))

  (define (cps-map expression continue environment)
    (define procedure (car expression))
    (define items (cadr expression))
    (define (iterate items values)
      (if (null? items)
        (continue (reverse values) environment)
        (let ((head (car items))
              (tail (cdr items)))
          (define (continue-after-item value environment)
            (iterate tail (cons value values)))
          (procedure (list head) continue-after-item environment))))
    (iterate items '()))

 (define (cps-for-each expression continue environment)
    (define procedure (car expression))
    (define items (cadr expression))
    (define (iterate items)
      (if (null? items)
        (continue '() environment)
        (let ((head (car items))
              (tail (cdr items)))
          (define (continue-after-item value environment)
            (iterate tail))
          (procedure (list head) continue-after-item environment))))
    (iterate items))

  (define (cps-call-cc expression continue environment)
    (define procedure (car expression))
    (define (continuation arguments dynamic-continue dynamic-environment)
      (continue (car arguments) environment))
    (procedure (list continuation) continue environment))

;
; read-eval-print
;

  (define (loop output environment)
    (define rollback environment)

    (define (error message qualifier)
      (display message)
      (loop qualifier rollback))

;
; functions
;

    (define (bind-variable variable value environment)
      (define binding (cons variable value))
      (cons binding environment))

    (define (bind-parameters parameters arguments environment)
      (if (symbol? parameters)
        (bind-variable parameters arguments environment)
        (if (pair? parameters)
          (let* 
            ((variable (car parameters))
             (value (car arguments ))
             (environment (bind-variable variable value environment)))
            (bind-parameters (cdr parameters) (cdr arguments) environment))
          environment)))

    (define (evaluate-sequence expressions continue environment)
      (define head (car expressions))
      (define tail (cdr expressions))
      (define (continue-with-sequence value environment)
        (evaluate-sequence tail continue environment))
      (if (null? tail)
        (evaluate head continue environment)
        (evaluate head continue-with-sequence environment)))

    (define (make-procedure parameters expressions environment)
      (lambda (arguments continue dynamic-environment)
        (define (continue-after-sequence value environment-after-sequence)
          (continue value dynamic-environment))
        (define lexical-environment (bind-parameters parameters arguments environment))
        (evaluate-sequence expressions continue-after-sequence lexical-environment)))

;
; evaluation functions
;

    (define (evaluate-application operator)
      (lambda operands
        (lambda (continue environment)
          (define (continue-after-operator procedure environment-after-operator)
            (define (evaluate-operands operands arguments environment)
              (define (continue-with-operands value environment-with-operands)
                (evaluate-operands (cdr operands) (cons value arguments) environment-with-operands))
              (if (null? operands)
                (procedure (reverse arguments) continue environment)
                (evaluate (car operands) continue-with-operands environment)))
            (evaluate-operands operands '() environment-after-operator))
          (evaluate operator continue-after-operator environment))))

    (define (evaluate-begin . expressions)
      (lambda (continue environment)
        (evaluate-sequence expressions continue environment)))

    (define (evaluate-define pattern . expressions)
      (lambda (continue environment)
        (define (continue-after-expression value environment-after-expression)
          (define binding (cons pattern value))
          (continue value (cons binding environment-after-expression)))
        (if (symbol? pattern)
          (evaluate (car expressions) continue-after-expression environment)
          (let* ((binding (cons (car pattern) '()))
                 (environment (cons binding environment))
                 (procedure (make-procedure (cdr pattern) expressions environment)))
              (set-cdr! binding procedure)
              (continue procedure environment)))))

    (define (evaluate-if predicate consequent . alternative)
      (lambda (continue environment)
        (define (continue-after-predicate boolean environment-after-predicate)
          (if (eq? boolean #f) 
            (if (null? alternative)
              (continue '() environment-after-predicate)
              (evaluate (car alternative) continue environment-after-predicate))
            (evaluate consequent continue environment-after-predicate)))
        (evaluate predicate continue-after-predicate environment)))

    (define (evaluate-lambda parameters . expressions)
      (lambda (continue environment)
        (continue (make-procedure parameters expressions environment) environment)))

    (define (evaluate-quote expression)
      (lambda (continue environment)
        (continue expression environment)))

    (define (evaluate-set! variable expression)
      (lambda (continue environment)
        (define (continue-after-expression value environment-after-expression)
          (define binding (assoc variable environment-after-expression))
          (if binding
            (set-cdr! binding value)
            (error "inaccessible variable: " variable))
          (continue value environment-after-expression))
        (evaluate expression continue-after-expression environment)))

    (define (evaluate-variable variable continue environment)
      (define binding (assoc variable environment))
      (if binding
        (continue (cdr binding) environment)
        (let ((native-value (meta-level-eval variable (interaction-environment))))
          (if (procedure? native-value)
            (continue (wrap-native-procedure native-value) environment)
            (continue native-value environment)))))

    (define (evaluate-while predicate . expressions)
      (lambda (continue environment)
        (define (iterate value environment)
          (define (continue-after-predicate boolean environment-after-predicate)
            (define (continue-after-sequence value environment-after-sequence)
              (iterate value environment))
            (if (eq? boolean #f)
              (continue value environment)
              (evaluate-sequence expressions continue-after-sequence environment)))
          (evaluate predicate continue-after-predicate environment))
        (iterate '() environment)))

;
; evaluator
;
  
    (define (evaluate expression continue environment)
      (cond
        ((symbol? expression)
         (evaluate-variable expression continue environment))
        ((pair? expression)
         (let ((operator (car expression))
               (operands (cdr expression)))
           ((apply
              (case operator
                ((begin)  evaluate-begin )
                ((define) evaluate-define)
                ((if)     evaluate-if    )
                ((lambda) evaluate-lambda)
                ((quote)  evaluate-quote ) 
                ((set!)   evaluate-set!  )
                ((while)  evaluate-while )
                (else     (evaluate-application operator))) operands) continue environment)))
        (else
          (continue expression environment))))

;
; read-eval-print
;

    (display output)
    (newline)
    (display ">>>")
    (evaluate (read) loop environment))

  (loop "cpSlip* version 2" (list (cons 'apply    cps-apply   ) 
                                  (cons 'map      cps-map     ) 
                                  (cons 'for-each cps-for-each) 
                                  (cons 'call-cc  cps-call-cc ))))